from django.db import models
from modelcluster.fields import ParentalKey, ParentalManyToManyField
from modelcluster.models import ClusterableModel, get_all_child_relations
from treebeard.mp_tree import MP_Node
from wagtail.fields import RichTextField, StreamField
from wagtail.models import COMMENTS_RELATION_NAME, Page, TranslatableMixin


class BaseTranslatableField:
    def __init__(self, field_name):
        self.field_name = field_name

    def get_field(self, model):
        return model._meta.get_field(self.field_name)

    def get_value(self, obj):
        return self.get_field(obj.__class__).value_from_object(obj)

    def is_editable(self, obj):
        """
        Returns True if the field is editable on the given object
        """
        return True

    def is_translated(self, obj):
        """
        Returns True if the value of this field on the given object should be
        extracted and submitted for translation
        """
        return False

    def is_synchronized(self, obj):
        """
        Returns True if the value of this field on the given object should be
        copied when translations are created/updated
        """
        return False

    def is_overridable(self, obj):
        """
        Returns True if the value of this field can be overridden. This is only
        applicable to fields that are synchronized
        """
        return self.is_synchronized(obj)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.field_name == other.field_name


class TranslatableField(BaseTranslatableField):
    """
    A field that should be translated whenever the original page changes
    """

    def is_translated(self, obj):
        return True

    def is_synchronized(self, source):
        field = self.get_field(source.__class__)

        # Child relations should all be synchronised before translation
        if isinstance(field, models.ManyToOneRel) and isinstance(
            field.remote_field, ParentalKey
        ):
            return True

        # We have a text field that has been cleared so we should mark it as synchronised
        if (
            isinstance(field, (RichTextField, models.TextField, models.CharField))
            and getattr(source, field.attname) == ""
        ):
            return True

        # Streamfields need to be re-synchronised before translation so
        # the structure and non-translatable content is copied over
        return isinstance(field, StreamField)

    def __repr__(self):
        return f"<TranslatableField {self.field_name}>"


class SynchronizedField(BaseTranslatableField):
    """
    A field that should always be kept in sync with the original page.
    """

    def __init__(self, field_name, overridable=True):
        super().__init__(field_name)
        self.overridable = overridable

    def is_synchronized(self, obj):
        return self.is_editable(obj)

    def is_overridable(self, obj):
        return self.is_synchronized(obj) and self.overridable

    def __repr__(self):
        return f"<SynchronizedField {self.field_name}>"


def get_translatable_fields(model):
    """
    Derives a list of translatable fields from the given model class.

    Arguments:
        model (Model class): The model class to derive translatable fields from.

    Returns:
        list[TranslatableField or SynchronizedField]: A list of TranslatableField and SynchronizedFields that were
            derived from the model.

    """
    # Note: If you update this, please update "docs/concept/translatable-fields-autogen.md"
    if hasattr(model, "translatable_fields"):
        return model.translatable_fields

    translatable_fields = []

    for field in model._meta.get_fields():
        # Ignore automatically generated IDs
        if isinstance(field, models.AutoField):
            continue

        # Ignore non-editable fields
        if not field.editable:
            continue

        # Ignore many to many fields (not supported yet)
        # TODO: Add support for these
        if isinstance(field, models.ManyToManyField):
            continue

        # Ignore fields defined by MP_Node mixin
        if issubclass(model, MP_Node) and field.name in ["path", "depth", "numchild"]:
            continue

        # Ignore some editable fields defined on Page
        if issubclass(model, Page) and field.name in [
            "go_live_at",
            "expire_at",
            "first_published_at",
            "content_type",
            "owner",
        ]:
            continue

        # URL, Email and choices fields are an exception to the rule below.
        # Text fields are translatable, but these are synchronised.
        if (
            isinstance(field, (models.URLField, models.EmailField))
            or isinstance(field, models.CharField)
            and field.choices
        ):
            translatable_fields.append(SynchronizedField(field.name))

        # Translatable text fields should be translatable
        elif isinstance(
            field, (StreamField, RichTextField, models.TextField, models.CharField)
        ):
            translatable_fields.append(TranslatableField(field.name))

        # Foreign keys to translatable models should be translated. Others should be synchronised
        elif isinstance(field, models.ForeignKey):
            # Ignore if this is a link to a parent model
            if isinstance(field, ParentalKey):
                continue

            # Ignore parent links
            if (
                isinstance(field, models.OneToOneField)
                and field.remote_field.parent_link
            ):
                continue

            # All FKs to translatable models should be translatable.
            # With the exception of pages that are special because we can localize them at runtime easily.
            # TODO: Perhaps we need a special type for pages where it links to the translation if availabe,
            # but falls back to the source if it isn't translated yet?
            # Note: This exact same decision was made for page chooser blocks in segments/extract.py
            if issubclass(field.related_model, TranslatableMixin) and not issubclass(
                field.related_model, Page
            ):
                translatable_fields.append(TranslatableField(field.name))
            else:
                translatable_fields.append(SynchronizedField(field.name))

        # Fields that support extracting segments are translatable
        elif hasattr(field, "get_translatable_segments"):
            translatable_fields.append(TranslatableField(field.name))

        else:
            # Everything else is synchronised
            translatable_fields.append(SynchronizedField(field.name))

    # Add child relations for clusterable models
    if issubclass(model, ClusterableModel):
        for child_relation in get_all_child_relations(model):
            # Ignore comments
            if (
                issubclass(model, Page)
                and child_relation.name == COMMENTS_RELATION_NAME
            ):
                continue

            if issubclass(child_relation.related_model, TranslatableMixin):
                translatable_fields.append(TranslatableField(child_relation.name))
            else:
                translatable_fields.append(SynchronizedField(child_relation.name))

    # Combine with any overrides defined on the model
    override_translatable_fields = getattr(model, "override_translatable_fields", [])

    if override_translatable_fields:
        override_translatable_fields = {
            field.field_name: field for field in override_translatable_fields
        }

        combined_translatable_fields = []
        for field in translatable_fields:
            if field.field_name in override_translatable_fields:
                combined_translatable_fields.append(
                    override_translatable_fields.pop(field.field_name)
                )
            else:
                combined_translatable_fields.append(field)

        if override_translatable_fields:
            combined_translatable_fields.extend(override_translatable_fields.values())

        return combined_translatable_fields

    else:
        return translatable_fields


def copy_synchronised_fields(source, target):
    """
    Copies data in synchronised fields from the source instance to the target instance.

    Note: Both instances must have the same model class

    Arguments:
        source (Model): The source instance to copy data from.
        target (Model): The target instance to copy data to.
    """
    for translatable_field in get_translatable_fields(source.__class__):
        if translatable_field.is_synchronized(source):
            field = translatable_field.get_field(target.__class__)

            if isinstance(field, models.ManyToOneRel) and isinstance(
                field.remote_field, ParentalKey
            ):
                # Use modelcluster's copy_child_relation for child relations

                if issubclass(field.related_model, TranslatableMixin):
                    # Get a list of the primary keys for the existing child objects
                    existing_pks_by_translation_key = {
                        child_object.translation_key: child_object.pk
                        for child_object in getattr(target, field.name).all()
                    }

                    # Copy the new child objects across (note this replaces existing ones)
                    child_object_map = source.copy_child_relation(field.name, target)

                    # Update locale of each child object and recycle PK
                    for (
                        _child_relation,
                        source_pk,
                    ), child_objects in child_object_map.items():
                        if source_pk is None:
                            # This is a list of the child objects that were never saved
                            for child_object in child_objects:
                                child_object.pk = existing_pks_by_translation_key.get(
                                    child_object.translation_key
                                )
                                child_object.locale = target.locale
                        else:
                            child_object = child_objects

                            child_object.pk = existing_pks_by_translation_key.get(
                                child_object.translation_key
                            )
                            child_object.locale = target.locale

                else:
                    source.copy_child_relation(field.name, target)

            elif isinstance(field, ParentalManyToManyField):
                parental_field = getattr(source, field.name)
                if hasattr(parental_field, "all"):
                    values = parental_field.all()
                    if values:
                        setattr(target, field.attname, values)

            else:
                # For all other fields, just set the attribute
                setattr(target, field.attname, getattr(source, field.attname))



from django.contrib.admin.utils import quote
from django.urls import reverse
from rest_framework import fields, serializers, viewsets
from wagtail.snippets.views.snippets import get_snippet_model_from_url_params


class BaseSnippetSerializer(serializers.ModelSerializer):
    id = fields.ReadOnlyField(source="pk")
    title = fields.SerializerMethodField()
    edit_url = fields.SerializerMethodField()

    def get_title(self, instance):
        return str(instance)

    def get_edit_url(self, instance):
        return reverse(
            f"wagtailsnippets_{instance._meta.app_label}_{instance._meta.model_name}:edit",
            args=[quote(instance.pk)],
        )


class SnippetViewSet(viewsets.ModelViewSet):
    def get_model(self):
        return get_snippet_model_from_url_params(
            self.kwargs["app_label"], self.kwargs["model_name"]
        )

    def get_queryset(self):
        return self.get_model()._default_manager.all()

    def get_serializer_class(self):
        model = self.get_model()

        return type(
            model.__name__ + "Serializer",
            (BaseSnippetSerializer,),
            {
                "Meta": type(
                    "Meta",
                    (object,),
                    {"model": model, "fields": ["id", "title", "edit_url"]},
                ),
            },
        )



import contextlib

from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.admin.utils import quote
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import transaction
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from django.views.generic import TemplateView
from django.views.generic.detail import SingleObjectMixin
from wagtail.admin.views.pages.utils import get_valid_next_url_from_request
from wagtail.models import Page
from wagtail.snippets.models import get_snippet_models
from wagtail.utils.version import get_main_version

from wagtail_localize.models import TranslationSource
from wagtail_localize.views.submit_translations import TranslationComponentManager


class UpdateTranslationsForm(forms.Form):
    publish_translations = forms.BooleanField(
        label=gettext_lazy("Publish immediately"),
        help_text=gettext_lazy(
            "This will apply the updates and publish immediately, before any new translations happen."
        ),
        required=False,
    )


class UpdateTranslationsView(SingleObjectMixin, TemplateView):
    template_name = "wagtail_localize/admin/update_translations.html"
    title = gettext_lazy("Update existing translations")

    def get_object(self):
        return get_object_or_404(
            TranslationSource, id=self.kwargs["translation_source_id"]
        )

    def get_title(self):
        return self.title

    def get_subtitle(self):
        return self.object.object_repr

    def get_form(self):
        if self.request.method == "POST":
            return UpdateTranslationsForm(self.request.POST)
        else:
            return UpdateTranslationsForm()

    def get_success_message(self):
        return _("Successfully updated translations for '{object}'").format(
            object=self.object.object_repr
        )

    def get_success_url(self):
        return get_valid_next_url_from_request(self.request)

    def get_default_success_url(self):
        instance = self.object.get_source_instance()

        if isinstance(instance, Page):
            return reverse("wagtailadmin_explore", args=[instance.get_parent().id])

        elif instance._meta.model in get_snippet_models():
            return reverse(
                f"wagtailsnippets_{instance._meta.app_label}_{instance._meta.model_name}:edit",
                args=[quote(instance.pk)],
            )

        elif "wagtail_localize.modeladmin" in settings.INSTALLED_APPS:
            return reverse(
                f"{instance._meta.app_label}_{instance._meta.model_name}_modeladmin_index"
            )

    def get_edit_url(self, instance):
        if isinstance(instance, Page):
            return reverse("wagtailadmin_pages:edit", args=[instance.id])

        elif instance._meta.model in get_snippet_models():
            return reverse(
                f"wagtailsnippets_{instance._meta.app_label}_{instance._meta.model_name}:edit",
                args=[quote(instance.pk)],
            )

        elif "wagtail_localize.modeladmin" in settings.INSTALLED_APPS:
            return reverse(
                f"{instance._meta.app_label}_{instance._meta.model_name}_modeladmin_edit",
                args=[quote(instance.pk)],
            )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "translations": [
                    {
                        "title": str(translation.get_target_instance()),
                        "locale": translation.target_locale,
                        "edit_url": self.get_edit_url(
                            translation.get_target_instance()
                        ),
                    }
                    for translation in self.object.translations.filter(
                        enabled=True
                    ).select_related("target_locale")
                ],
                "last_sync_date": self.object.last_updated_at,
                "form": self.get_form(),
                "next_url": self.get_success_url(),
                "back_url": self.get_success_url() or self.get_default_success_url(),
                "components": self.components,
                "wagtail_version": get_main_version(),
            }
        )
        return context

    def post(self, request, **kwargs):
        form = self.get_form()

        if form.is_valid() and self.components.is_valid():
            return self.form_valid(form)

        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    @transaction.atomic
    def form_valid(self, form):
        self.object.update_from_db()

        enabled_translations = self.object.translations.filter(enabled=True)

        if form.cleaned_data["publish_translations"]:
            for translation in enabled_translations.select_related("target_locale"):
                with contextlib.suppress(ValidationError):
                    translation.save_target(user=self.request.user, publish=True)
        else:
            for translation in enabled_translations.select_related(
                "source", "target_locale"
            ):
                with contextlib.suppress(ValidationError):
                    translation.source.update_target_view_restrictions(
                        translation.target_locale
                    )

        self.components.save(
            self.object,
            sources_and_translations={self.object: list(enabled_translations)},
        )

        # TODO: Button that links to page in translations report when we have it
        messages.success(self.request, self.get_success_message())

        return redirect(self.get_success_url() or self.get_default_success_url())

    def dispatch(self, request, *args, **kwargs):
        if not request.user.has_perms(["wagtail_localize.submit_translation"]):
            raise PermissionDenied

        self.object = self.get_object()
        self.components = TranslationComponentManager.from_request(
            self.request, source_object_instance=self.object
        )
        return super().dispatch(request, *args, **kwargs)



# Generated by Django 3.1 on 2020-08-28 10:00

import django.db.models.deletion

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wagtail_localize", "0009_stringtranslation_errors"),
    ]

    operations = [
        migrations.CreateModel(
            name="OverridableSegment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("order", models.PositiveIntegerField()),
                ("data_json", models.TextField()),
                (
                    "context",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="wagtail_localize.translationcontext",
                    ),
                ),
                (
                    "source",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="wagtail_localize.translationsource",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



import contextlib

from collections import defaultdict

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import transaction
from wagtail.models import DraftStateMixin, Page

from wagtail_localize.models import Translation, TranslationSource


class TranslationCreator:
    """
    A class that provides a create_translations method.

    Call create_translations for each object you want to translate and this will submit
    that object and any dependencies as well.

    This class will track the objects that have already submitted so an object doesn't
    get submitted twice.
    """

    def __init__(self, user, target_locales):
        self.user = user
        self.target_locales = target_locales
        self.seen_objects = set()
        self.mappings = defaultdict(list)

    def create_translations(self, instance, include_related_objects=True):
        if isinstance(instance, Page):
            instance = instance.specific

        if instance.translation_key in self.seen_objects:
            return
        self.seen_objects.add(instance.translation_key)

        source, created = TranslationSource.get_or_create_from_instance(instance)

        # Add related objects
        # Must be before translation records or those translation records won't be able to create
        # the objects because the dependencies haven't been created
        if include_related_objects:
            for related_object_segment in source.relatedobjectsegment_set.all():
                related_instance = related_object_segment.object.get_instance(
                    instance.locale
                )

                # Limit to one level of related objects, since this could potentially pull in a lot of stuff
                self.create_translations(
                    related_instance, include_related_objects=False
                )

        # Support disabling the out of the box translation mode.
        # The value set on the model takes precedence over the global setting.
        if hasattr(instance, "localize_default_translation_mode"):
            translation_mode = instance.localize_default_translation_mode
        else:
            translation_mode = getattr(
                settings, "WAGTAIL_LOCALIZE_DEFAULT_TRANSLATION_MODE", "synced"
            )
        translation_enabled = translation_mode == "synced"

        # Set up translation records
        for target_locale in self.target_locales:
            # Skip if target_locale is the same as source locale
            if target_locale == source.locale:
                continue

            # Create translation if it doesn't exist yet, re-enable if translation was disabled
            # Note that the form won't show this locale as an option if the translation existed
            # in this language, so this shouldn't overwrite any unmanaged translations.
            translation, created = Translation.objects.update_or_create(
                source=source,
                target_locale=target_locale,
                defaults={"enabled": translation_enabled},
            )

            self.mappings[source].append(translation)

            # Determine whether to publish the translation.
            if getattr(settings, "WAGTAILLOCALIZE_SYNC_LIVE_STATUS_ON_TRANSLATE", True):
                publish = getattr(instance, "live", True)
            else:
                # If the model can't be saved as a draft, then we have to publish it
                publish = not isinstance(instance, DraftStateMixin)

            with contextlib.suppress(ValidationError):
                translation.save_target(user=self.user, publish=publish)


@transaction.atomic
def translate_object(instance, locales, components=None, user=None):
    """
    Translates the given object into the given locales.
    """
    translator = TranslationCreator(user, locales)
    translator.create_translations(instance)

    if components is not None:
        components.save(translator, sources_and_translations=translator.mappings)


@transaction.atomic
def translate_page_subtree(page_id, locales, components, user):
    """
    Translates the given page's subtree into the given locales.

    Note that the page itself must already be translated.

    Note: Page must be passed by ID since this function may be called with an async worker and pages can't be reliably pickled.
          See: https://github.com/wagtail/wagtail/pull/5998
    """
    page = Page.objects.get(id=page_id)

    translator = TranslationCreator(user, locales)

    def _walk(current_page):
        for child_page in current_page.get_children():
            translator.create_translations(child_page)

            if child_page.numchild:
                _walk(child_page)

    _walk(page)

    if components is not None:
        components.save(translator, sources_and_translations=translator.mappings)



import os


# Windows has a different strftime format for dates without leading 0
# https://stackoverflow.com/questions/904928/python-strftime-date-without-leading-0
DATE_FORMAT = "%#d %B %Y" if os.name == "nt" else "%-d %B %Y"



from collections import defaultdict

from django.apps import apps
from django.db import models
from wagtail import blocks
from wagtail.fields import RichTextField, StreamField
from wagtail.rich_text import RichText

from wagtail_localize.strings import restore_strings

from .types import OverridableSegmentValue, StringSegmentValue


def unquote_path_component(text):
    """
    Removes quotes around a quoted path component, and unescapes any special characters.
    """
    if text[0] != "'" or text[-1] != "'":
        raise ValueError("value must be a quoted string")

    return text[1:-1].replace("\\'", "'").replace("\\\\", "\\")


def organise_template_segments(segments):
    """
    Organises the segments for a RichTextField or RichTextBlock to prepare them for recombining.

    This takes a single TemplateValue segment, and zero or more StringValueSegments/OverridableValueSegments.
    It detects the TemplateValue and groups the OverridableValueSegments with their StringValueSegments.

    Arguments:
        segments (list[StringValueSegment, TemplateValueSegment, or OverridableValueSegment]): The segments to organise.

    Returns:
        tuple[string, string, list[tuple[string, dict]]]: Returns a 3-tuple. The first component is the template format,
            the second component is the template itself as a string, the third component is a list of 2-tuples that
            represent translated strings to be reinserted into the template. These 2-tuples contain the string itself
            and a dict of HTML attributes to insert back into the string.
    """
    # The first segment is always the template, followed by the texts in order of their position

    segments.sort(key=lambda segment: segment.order)
    template = segments[0]
    xrefs = {
        unquote_path_component(segment.path): segment.data
        for segment in segments
        if isinstance(segment, OverridableSegmentValue) and segment.data
    }

    def translate_href(attrs):
        """Update href in segments with their translated values."""
        if attrs:
            for val in attrs.values():
                if val and "href" in val and val["href"] in xrefs:
                    val["href"] = xrefs[val["href"]]
        return attrs

    return (
        template.format,
        template.template,
        [
            (segment.string, translate_href(segment.attrs))
            for segment in segments[1:]
            if isinstance(segment, StringSegmentValue)
        ],
    )


def handle_related_object(related_model, src_locale, tgt_locale, segments):
    """
    Returns the instance of the related object that is referenced by the given segments.

    This is called when inserting segments into a ForeignKey or ChooserBlock.

    Raises:
        Model.DoesNotExist: If the segment is a RelatedObjectValue and the related object doesn't exist in the target locale.
        Model.DoesNotExist: If the segment is a OverridableSegmentValue and the referenced instance no longer exists.

    Returns:
        Model: The referenced instance.
    """
    if len(segments) > 1:
        raise ValueError(
            f"Related object field can only have a single segment. Found {len(segments)}"
        )

    segment = segments[0]

    if isinstance(segment, OverridableSegmentValue):
        return related_model.objects.get(pk=segment.data)

    else:
        # Assume it's a RelatedObjectValue
        return segment.get_instance(tgt_locale)


class StreamFieldSegmentsWriter:
    """
    A helper class to help traverse StreamField values and insert translated segments.
    """

    def __init__(self, field, src_locale, tgt_locale):
        """
        Initialises a StreamFieldSegmentsWriter.

        Args:
            field (StreamField): The StreamField to extract segments from.
            src_locale (Locale): The source locale of the segments.
            tgt_locale (Locale): The locale of the segment translations.
        """
        self.field = field
        self.src_locale = src_locale
        self.tgt_locale = tgt_locale

    def handle_block(self, block_type, block_value, segments):
        # Need to check if the app is installed before importing EmbedBlock
        # See: https://github.com/wagtail/wagtail-localize/issues/309
        if apps.is_installed("wagtail.embeds"):
            from wagtail.embeds.blocks import EmbedBlock, EmbedValue

            if isinstance(block_type, EmbedBlock):
                if len(segments) > 1:
                    raise ValueError(
                        f"EmbedBlock can only have a single segment. Found {len(segments)}"
                    )

                segment = segments[0]

                if isinstance(segment, OverridableSegmentValue):
                    return EmbedValue(segment.data)

        if hasattr(block_type, "restore_translated_segments"):
            return block_type.restore_translated_segments(block_value, segments)

        elif isinstance(
            block_type,
            (blocks.CharBlock, blocks.TextBlock, blocks.URLBlock, blocks.EmailBlock),
        ):
            if len(segments) > 1:
                raise ValueError(
                    f"TextBlock/CharBlock can only have a single segment. Found {len(segments)}"
                )

            segment = segments[0]

            if isinstance(segment, OverridableSegmentValue):
                return segment.data
            else:
                # Assume it's a StringSegmentValue
                return segment.render_text()

        elif isinstance(block_type, blocks.RichTextBlock):
            segment_format, template, strings = organise_template_segments(segments)
            if segment_format != "html":
                raise ValueError(
                    f"RichTextBlock can only contain HTML segments. Found {segment_format}"
                )
            return RichText(restore_strings(template, strings))

        elif isinstance(block_type, blocks.ChooserBlock):
            return self.handle_related_object_block(block_value, segments)

        elif isinstance(block_type, blocks.StructBlock):
            return self.handle_struct_block(block_value, segments)

        elif isinstance(block_type, blocks.ListBlock):
            return self.handle_list_block(block_value, segments)

        elif isinstance(block_type, blocks.StreamBlock):
            return self.handle_stream_block(block_value, segments)

        else:
            raise Exception(
                f"Unrecognised StreamField block type '{block_type.__class__.__name__}'. Have you implemented restore_translated_segments() on this class?"
            )

    def handle_related_object_block(self, related_object, segments):
        return handle_related_object(
            related_object.__class__, self.src_locale, self.tgt_locale, segments
        )

    def handle_struct_block(self, struct_block, segments):
        segments_by_field = defaultdict(list)

        for segment in segments:
            field_name, segment = segment.unwrap()
            segments_by_field[field_name].append(segment)

        for field_name, segments in segments_by_field.items():
            if segments:
                block_type = struct_block.block.child_blocks[field_name]
                block_value = struct_block[field_name]
                struct_block[field_name] = self.handle_block(
                    block_type, block_value, segments
                )

        return struct_block

    def handle_list_block(self, list_block, segments):
        segments_by_block = defaultdict(list)

        for segment in segments:
            block_uuid, segment = segment.unwrap()
            segments_by_block[block_uuid].append(segment)

        for block_index, block in enumerate(list_block.bound_blocks):
            block_segments = segments_by_block[block.id]
            list_block.bound_blocks[block_index].value = self.handle_block(
                block.block, block.value, block_segments
            )

        return list_block

    def get_stream_block_child_data(self, stream_block, block_uuid):
        for stream_child in stream_block:
            if stream_child.id == block_uuid:
                return stream_child

    def handle_stream_block(self, stream_block, segments):
        segments_by_block = defaultdict(list)

        for segment in segments:
            block_uuid, segment = segment.unwrap()
            segments_by_block[block_uuid].append(segment)

        for block_uuid, segments in segments_by_block.items():
            block = self.get_stream_block_child_data(stream_block, block_uuid)
            block.value = self.handle_block(block.block, block.value, segments)

        return stream_block


def ingest_segments(original_obj, translated_obj, src_locale, tgt_locale, segments):
    """
    Ingests translated segments into a translated instance.

    Args:
        original_obj (Model): The original instance that the segments were extracted from.
        translated_obj (Model): The translated instance that we are ingesting segments into.
        src_local (Locale): The locale of the source instance.
        tgt_locale (Locale): The locale of the translated instance
        segments (list[StringSegmentValue, TemplateSegmentValue, RelatedObjectSegmentValue, or OverridableSegmentValue]):
            The segment values to ingest
    """
    # Get segments by field name
    segments_by_field_name = defaultdict(list)
    for segment in segments:
        field_name, segment = segment.unwrap()
        segments_by_field_name[field_name].append(segment)

    for field_name, field_segments in segments_by_field_name.items():
        field = translated_obj.__class__._meta.get_field(field_name)

        if hasattr(field, "restore_translated_segments"):
            value = field.value_from_object(original_obj)
            new_value = field.restore_translated_segments(value, field_segments)
            setattr(translated_obj, field_name, new_value)

        elif isinstance(field, StreamField):
            data = field.value_from_object(original_obj)
            StreamFieldSegmentsWriter(
                field, src_locale, tgt_locale
            ).handle_stream_block(data, field_segments)
            setattr(translated_obj, field_name, data)

        elif isinstance(field, RichTextField):
            segment_format, template, strings = organise_template_segments(
                field_segments
            )
            if segment_format != "html":
                raise ValueError(
                    f"RichTextField can only contain HTML segments. Found {segment_format}"
                )
            html = restore_strings(template, strings)
            setattr(translated_obj, field_name, html)

        elif isinstance(field, (models.TextField, models.CharField)):
            if len(field_segments) > 1:
                raise ValueError(
                    f"TextField/CharField can only have a single segment. Found {len(field_segments)}"
                )

            segment = field_segments[0]

            if isinstance(segment, OverridableSegmentValue):
                setattr(translated_obj, field_name, segment.data)
            else:
                # Assume it's a StringSegmentValue
                setattr(translated_obj, field_name, segment.render_text())

        elif isinstance(field, models.ForeignKey):
            related_translated = handle_related_object(
                field.related_model, src_locale, tgt_locale, field_segments
            )
            setattr(translated_obj, field_name, related_translated)

        elif isinstance(field, (models.ManyToOneRel)):
            original_manager = getattr(original_obj, field_name)
            translated_manager = getattr(translated_obj, field_name)

            segments_by_child = defaultdict(list)

            for segment in field_segments:
                child_translation_key, segment = segment.unwrap()
                segments_by_child[child_translation_key].append(segment)

            for child_translation_key, child_segments in segments_by_child.items():
                # The child objects must be synchronised before calling this function, so we
                # can assume that both exist
                original_child_object = original_manager.get(
                    translation_key=child_translation_key
                )
                translated_child_object = translated_manager.get(
                    translation_key=child_translation_key
                )

                ingest_segments(
                    original_child_object,
                    translated_child_object,
                    src_locale,
                    tgt_locale,
                    child_segments,
                )



from django import VERSION as DJANGO_VERSION
from django.apps import apps
from django.conf import settings
from django.contrib.admin.utils import unquote
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ImproperlyConfigured
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.urls import NoReverseMatch, reverse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from wagtail.models import Locale, TranslatableMixin
from wagtail.utils.version import get_main_version
from wagtail_modeladmin.views import (
    ChooseParentView,
    CreateView,
    DeleteView,
    EditView,
    HistoryView,
    IndexView,
    InspectView,
)

from wagtail_localize.models import Translation
from wagtail_localize.views import edit_translation
from wagtail_localize.views.submit_translations import SubmitTranslationView


class TranslatableViewMixin:
    def __init__(self, *args, **kwargs):
        self.locale = None
        super().__init__(*args, **kwargs)

        if "wagtail_localize.modeladmin" not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured(
                'To use the TranslatableView class "wagtail_localize.modeladmin" '
                "must be added to your INSTALLED_APPS setting."
            )

        if not issubclass(self.model, TranslatableMixin):
            raise ImproperlyConfigured(
                f"Model `{self.model}` used in translatable view `{self.__class__}` "
                f"must subclass the `{TranslatableMixin}` class."
            )

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        if getattr(self, "instance", None):
            self.locale = self.instance.locale
        if "locale" in request.GET:
            self.locale = get_object_or_404(Locale, language_code=request.GET["locale"])
        if not self.locale:
            self.locale = Locale.get_active()
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["locale"] = self.locale
        context["translations"] = []
        context["wagtail_version"] = get_main_version()
        return context


class TranslatableIndexView(TranslatableViewMixin, IndexView):
    def get_filters(self, request):
        filters = super().get_filters(request)
        # Update the 'lookup_params' part of the filters tuple to filter objects
        # using the currently active locale
        if DJANGO_VERSION >= (5, 0):
            # modeladmin IndexView needs an iterable in build_q_object_from_lookup_parameters
            filters[2]["locale"] = [self.locale.id]
        else:
            filters[2]["locale"] = self.locale.id
        return filters

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["translations"] = [
            {
                "locale": locale,
                "url": self.index_url + "?locale=" + locale.language_code,
            }
            for locale in Locale.objects.exclude(id=self.locale.id)
        ]
        return context


class TranslatableCreateView(TranslatableViewMixin, CreateView):
    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["instance"].locale = self.locale
        return kwargs

    def get_success_url(self):
        return self.index_url + "?locale=" + self.locale.language_code

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["translations"] = [
            {
                "locale": locale,
                "url": self.create_url + "?locale=" + locale.language_code,
            }
            for locale in Locale.objects.exclude(id=self.locale.id)
        ]
        return context


class TranslatableEditView(TranslatableViewMixin, EditView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Attempt to find the translation for the edited instance or set to None
        self.translation = Translation.objects.filter(
            source__object_id=self.instance.translation_key,
            target_locale_id=self.instance.locale_id,
        ).first()

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        # Check if the user has clicked the "Start Synced translation" menu item
        if (
            request.method == "POST"
            and "localize-restart-translation" in request.POST
            and self.translation
            and not self.translation.enabled
        ):
            return edit_translation.restart_translation(
                request, self.translation, self.instance
            )

        # Overrides the edit view if the object is translatable and the target of a translation
        if self.translation and self.translation.enabled:
            return edit_translation.edit_translation(
                request, self.translation, self.instance
            )

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["translation"] = self.translation
        context["translations"] = [
            {
                "locale": translation.locale,
                "url": self.url_helper.get_action_url("edit", translation.pk)
                + "?locale="
                + translation.locale.language_code,
            }
            for translation in self.instance.get_translations().select_related("locale")
        ]
        return context


class TranslatableInspectView(TranslatableViewMixin, InspectView):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["translations"] = [
            {
                "locale": translation.locale,
                "url": self.url_helper.get_action_url("inspect", translation.pk)
                + "?locale="
                + translation.locale.language_code,
            }
            for translation in self.instance.get_translations().select_related("locale")
        ]
        return context


class TranslatableDeleteView(TranslatableViewMixin, DeleteView):
    pass


class TranslatableHistoryView(TranslatableViewMixin, HistoryView):
    pass


class TranslatableChooseParentView(TranslatableViewMixin, ChooseParentView):
    pass


class SubmitModelAdminTranslationView(SubmitTranslationView):
    def get_title(self):
        return _("Translate {model_name}").format(
            model_name=self.object._meta.verbose_name
        )

    def get_object(self):
        try:
            model = apps.get_model(self.kwargs["app_label"], self.kwargs["model_name"])
        except LookupError as err:
            raise Http404 from err
        if not issubclass(model, TranslatableMixin):
            raise Http404
        return get_object_or_404(model, pk=unquote(self.kwargs["pk"]))

    def get_default_success_url(self, translated_object=None):
        pk = translated_object.pk if translated_object else self.kwargs["pk"]
        try:
            return reverse(
                "{app_label}_{model_name}_modeladmin_edit".format(**self.kwargs),
                args=[pk],
            )
        except NoReverseMatch as err:
            raise Http404 from err

    def get_success_message(self, locales):
        return _(
            "The {model_name} '{object}' was successfully submitted for translation into {locales}"
        ).format(
            model_name=self.object._meta.verbose_name,
            object=str(self.object),
            locales=locales,
        )



from django import template


register = template.Library()


@register.inclusion_tag(
    "wagtail_localize/admin/_language_switch.html", takes_context=True
)
def admin_language_switch(context):
    page = context["page"]

    return {
        "page": page,
        "language_count": page.get_translations().count(),
    }



import warnings

import requests

from wagtail_localize.strings import StringValue

from .base import BaseMachineTranslator


SUPPORTED_FORMALITY_OPTIONS = {"default", "prefer_less", "prefer_more"}


def language_code(code, *, is_target=False):
    # DeepL supports targeting Brazilian Portuguese and requires to specifically request American or British English.
    # @see https://www.deepl.com/en/docs-api/translate-text/translate-text
    upper_code = code.upper()
    if is_target and upper_code in ["PT-PT", "PT-BR", "EN-US", "EN-GB"]:
        return upper_code

    return upper_code.split("-")[0]


class DeepLTranslator(BaseMachineTranslator):
    display_name = "DeepL"

    def get_api_endpoint(self):
        if self.options.get("AUTH_KEY", "").endswith(":fx"):
            return "https://api-free.deepl.com/v2/translate"
        return "https://api.deepl.com/v2/translate"

    def translate(self, source_locale, target_locale, strings):
        parameters = {
            "auth_key": self.options["AUTH_KEY"],
            "text": [string.data for string in strings],
            "tag_handling": "xml",
            "source_lang": language_code(source_locale.language_code),
            "target_lang": language_code(target_locale.language_code, is_target=True),
        }

        if self.options.get("FORMALITY"):
            if self.options["FORMALITY"] in SUPPORTED_FORMALITY_OPTIONS:
                parameters["formality"] = self.options["FORMALITY"]
            else:
                warnings.warn(
                    f"Unsupported formality option '{self.options['FORMALITY']}'. Supported options are: {', '.join(SUPPORTED_FORMALITY_OPTIONS)}"
                )

        response = requests.post(
            self.get_api_endpoint(),
            parameters,
            timeout=30,
        )

        return {
            string: StringValue(translation["text"])
            for string, translation in zip(strings, response.json()["translations"])
        }

    def can_translate(self, source_locale, target_locale):
        return language_code(source_locale.language_code) != language_code(
            target_locale.language_code, is_target=True
        )



from wagtail.admin.forms.models import WagtailAdminModelForm


LOCALE_COMPONENTS = []


def get_locale_components():
    return LOCALE_COMPONENTS


def register_locale_component(*, heading, help_text=None, required=False):
    def _wrapper(model):
        if model not in LOCALE_COMPONENTS:
            LOCALE_COMPONENTS.append(
                {
                    "heading": heading,
                    "help_text": help_text,
                    "required": required,
                    "model": model,
                    "slug": model._meta.db_table,
                }
            )
            LOCALE_COMPONENTS.sort(key=lambda x: x["model"]._meta.verbose_name)

        return model

    return _wrapper


class LocaleComponentModelForm(WagtailAdminModelForm):
    def validate_with_locale(self, locale):
        """
        Validates the locale component against the given locale.
        """



import json

import requests

from wagtail_localize.machine_translators.base import BaseMachineTranslator
from wagtail_localize.strings import StringValue


class LibreTranslator(BaseMachineTranslator):
    """
    A machine translator that uses the LibreTranslate API.

    API Documentation:
        https://libretranslate.com/docs/
    """

    display_name = "LibreTranslate"

    def get_api_endpoint(self):
        return self.options["LIBRETRANSLATE_URL"]

    def get_timeout(self):
        return self.options.get("TIMEOUT", 10)

    def language_code(self, code):
        return code.split("-")[0]

    def translate(self, source_locale, target_locale, strings):
        translations = [item.data for item in list(strings)]
        response = requests.post(
            self.get_api_endpoint() + "/translate",
            data=json.dumps(
                {
                    "q": translations,
                    "source": self.language_code(source_locale.language_code),
                    "target": self.language_code(target_locale.language_code),
                    "api_key": self.options["API_KEY"],
                    "format": "html",
                }
            ),
            headers={"Content-Type": "application/json"},
            timeout=self.get_timeout(),
        )
        response.raise_for_status()

        return {
            string: StringValue(translation)
            for string, translation in zip(strings, response.json()["translatedText"])
        }

    def can_translate(self, source_locale, target_locale):
        return self.language_code(source_locale.language_code) != self.language_code(
            target_locale.language_code
        )



import django_filters

from django.contrib.contenttypes.models import ContentType
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy
from django_filters.constants import EMPTY_VALUES
from django_filters.fields import ModelChoiceField
from modelcluster.fields import ParentalKey
from wagtail import VERSION as WAGTAIL_VERSION
from wagtail.admin.filters import WagtailFilterSet
from wagtail.admin.views.reports import ReportView
from wagtail.coreutils import get_content_languages
from wagtail.models import get_translatable_models

from wagtail_localize.models import Translation


class SourceTitleFilter(django_filters.CharFilter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        return qs.filter(source__object_repr__icontains=value)


class ContentTypeModelChoiceField(ModelChoiceField):
    def label_from_instance(self, obj):
        return capfirst(obj.name)


class ContentTypeModelChoiceFilter(django_filters.ModelChoiceFilter):
    field_class = ContentTypeModelChoiceField

    def get_translatable_models(self, descendant_of=None):
        """
        Returns a list of model classes that are usable in this filter.

        When descendant_of is set, the models will be filtered to only include those that inherit from this model
        (includes the model itself if it's translatable)
        """
        translatable_models = get_translatable_models(include_subclasses=True)

        # Remove child models
        translatable_models = [
            model
            for model in translatable_models
            if not any(
                isinstance(field, ParentalKey) for field in model._meta.get_fields()
            )
        ]

        # Filter content types to only include those based on the provided type
        if descendant_of is not None:
            translatable_models = [
                model
                for model in translatable_models
                if issubclass(model, descendant_of)
            ]

        return translatable_models

    def get_queryset(self, request):
        """
        Returns QuerySet of ContentTypes to show in the filter.

        Called from django_filters.
        """
        translatable_models = self.get_translatable_models()

        # Return QuerySet of content types
        return ContentType.objects.filter(
            id__in=[
                content_type.id
                for content_type in ContentType.objects.get_for_models(
                    *translatable_models
                ).values()
            ]
        ).order_by("model")

    def filter(self, qs, value):
        """
        Filters the QuerySet to only include translations for the select content type and any descendant content types

        Called from django_filters.
        """
        if value in EMPTY_VALUES:
            return qs

        # Filter by content types that inherit from this one as well
        translatable_models = self.get_translatable_models(
            descendant_of=value.model_class()
        )
        return qs.filter(
            source__specific_content_type_id__in=[
                content_type.id
                for content_type in ContentType.objects.get_for_models(
                    *translatable_models
                ).values()
            ]
        )


def _get_locale_choices():
    return list(get_content_languages().items())


class LocaleFilter(django_filters.ChoiceFilter):
    def filter(self, qs, value):
        if not value or value == self.null_value:
            return qs

        return qs.filter(**{self.field_name + "__language_code": value})


class TranslationsReportFilterSet(WagtailFilterSet):
    content_type = ContentTypeModelChoiceFilter(
        field_name="source__specific_content_type", label=gettext_lazy("Content type")
    )
    source_title = SourceTitleFilter(label=gettext_lazy("Source title"))
    source_locale = LocaleFilter(
        field_name="source__locale",
        choices=_get_locale_choices,
        empty_label=None,
        null_label=gettext_lazy("All"),
        null_value="all",
    )
    target_locale = LocaleFilter(
        choices=_get_locale_choices,
        empty_label=None,
        null_label=gettext_lazy("All"),
        null_value="all",
    )

    class Meta:
        model = Translation
        fields = ["content_type", "source_title", "source_locale", "target_locale"]


def _adapt_wagtail_report_attributes(cls):
    """
    A class decorator that adapts ReportView-derived classes for compatibility
    with multiple versions of Wagtail by conditionally assigning attributes
    based on the Wagtail version. This includes setting appropriate titles,
    and adjusting template names and URL names for AJAX support since Wagtail 6.2.

    Attributes adjusted:
    - `title` or `page_title` based on Wagtail version for the display name of the report.
    - For Wagtail 6.2 and above, additional attributes like `results_template_name`,
      `index_results_url_name`, and `index_url_name` are set to support AJAX updates
      and utilize the `wagtail.admin.ui.tables` framework.
    """
    if WAGTAIL_VERSION < (6, 2):
        cls.title = gettext_lazy("Translations")
    else:
        # The `title` attr was **changed** to `page_title` in Wagtail 6.2
        cls.page_title = gettext_lazy("Translations")
        # The `results_template_name` attr was **added** in Wagtail 6.2
        # to support updating the listing via AJAX upon filtering and
        # to allow the use of the `wagtail.admin.ui.tables` framework.
        cls.results_template_name = (
            "wagtail_localize/admin/translations_report_results.html"
        )
        # The `index_results_url_name` attr was **added** in Wagtail 6.2
        # to support updating the listing via AJAX upon filtering.
        cls.index_results_url_name = "wagtail_localize:translations_report_results"
        cls.index_url_name = "wagtail_localize:translations_report"
    return cls


@_adapt_wagtail_report_attributes
class TranslationsReportView(ReportView):
    template_name = "wagtail_localize/admin/translations_report.html"
    header_icon = "site"

    filterset_class = TranslationsReportFilterSet

    def get_queryset(self):
        return Translation.objects.all()



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from modelcluster.fields import ParentalKey
from wagtail import blocks
from wagtail.fields import RichTextField, StreamField
from wagtail.models import Page, TranslatableMixin

from wagtail_localize.segments import (
    OverridableSegmentValue,
    RelatedObjectSegmentValue,
    StringSegmentValue,
    TemplateSegmentValue,
)

from ..fields import get_translatable_fields
from ..strings import extract_strings


def quote_path_component(text):
    """
    Puts quotes around the path compoenents, and escapes any special characters.
    """
    return "'" + text.replace("\\", "\\\\").replace("'", "\\'") + "'"


class StreamFieldSegmentExtractor:
    """
    A helper class to help traverse StreamField values and extract segments.
    """

    def __init__(self, field, include_overridables=False):
        """
        Initialises a StreamFieldSegmentExtractor.

        Args:
            field (StreamField): The StreamField to extract segments from.
            include_overridables (boolean, optional): Set this to True to extract overridable segments too.
        """
        self.field = field
        self.include_overridables = include_overridables

    def handle_block(self, block_type, block_value, raw_value=None):
        # Need to check if the app is installed before importing EmbedBlock
        # See: https://github.com/wagtail/wagtail-localize/issues/309
        if apps.is_installed("wagtail.embeds"):
            from wagtail.embeds.blocks import EmbedBlock

            if isinstance(block_type, EmbedBlock):
                if self.include_overridables:
                    return [OverridableSegmentValue("", block_value.url)]
                else:
                    return []

        if hasattr(block_type, "get_translatable_segments"):
            return block_type.get_translatable_segments(block_value)

        elif isinstance(block_type, (blocks.URLBlock, blocks.EmailBlock)):
            if self.include_overridables:
                return [OverridableSegmentValue("", block_value)]
            else:
                return []

        elif isinstance(
            block_type, (blocks.CharBlock, blocks.TextBlock, blocks.BlockQuoteBlock)
        ):
            return [StringSegmentValue("", block_value)]

        elif isinstance(block_type, blocks.RichTextBlock):
            template, strings = extract_strings(block_value.source)

            # Find all unique href values
            hrefs = set()
            for _string, attrs in strings:
                for tag_attrs in attrs.values():
                    if "href" in tag_attrs:
                        hrefs.add(tag_attrs["href"])

            ret = (
                [TemplateSegmentValue("", "html", template, len(strings))]
                + [
                    StringSegmentValue("", string, attrs=attrs)
                    for string, attrs in strings
                ]
                + [
                    OverridableSegmentValue(quote_path_component(href), href)
                    for href in sorted(hrefs)
                ]
            )
            return ret

        elif isinstance(block_type, blocks.ChooserBlock):
            return self.handle_related_object_block(block_value)

        elif isinstance(block_type, blocks.StructBlock):
            return self.handle_struct_block(block_value, raw_value=raw_value)

        elif isinstance(block_type, blocks.ListBlock):
            return self.handle_list_block(block_value, raw_value=raw_value)

        elif isinstance(block_type, blocks.StreamBlock):
            return self.handle_stream_block(block_value)

        # Ignore everything else
        return []

    def handle_related_object_block(self, related_object):
        if related_object is None:
            return []

        # All FKs to translatable models should be translatable.
        # With the exception of pages that are special because we can localize them at runtime easily.
        # TODO: Perhaps we need a special type for pages where it links to the translation if availabe,
        # but falls back to the source if it isn't translated yet?
        # Note: This exact same decision was made for regular foreign keys in fields.py
        if isinstance(related_object, TranslatableMixin) and not isinstance(
            related_object, Page
        ):
            return [RelatedObjectSegmentValue.from_instance("", related_object)]
        else:
            return [OverridableSegmentValue("", related_object.pk)]

    def handle_struct_block(self, struct_block, raw_value=None):
        segments = []

        for field_name, block_value in struct_block.items():
            block_type = struct_block.block.child_blocks[field_name]
            try:
                block_raw_value = raw_value["value"].get(field_name)
            except (KeyError, TypeError):
                # e.g. raw_value is None, or is that from chooser
                block_raw_value = None
            segments.extend(
                segment.wrap(field_name)
                for segment in self.handle_block(
                    block_type, block_value, raw_value=block_raw_value
                )
            )

        return segments

    def handle_list_block(self, list_block, raw_value=None):
        segments = []
        if raw_value is not None:
            # Wagtail 2.16 changes ListBlock values to be ListValue objects (i.e. {'value': '', 'id': ''})
            # and will automatically convert from the simple list format used before. However that requires
            # the block to be saved. bound_blocks will return ListValue objects, so we need to check that the
            # stored value is the new format before extracting segments, otherwise the block ids will continue
            # to change.
            has_block_format = False
            if (
                isinstance(raw_value, dict)
                and "value" in raw_value
                and len(raw_value["value"]) > 0
            ):
                has_block_format = list_block.list_block._item_is_in_block_format(
                    raw_value["value"][0]
                )
            elif isinstance(raw_value, list) and len(raw_value) > 0:
                has_block_format = list_block.list_block._item_is_in_block_format(
                    raw_value[0]
                )
            if has_block_format:
                for index, block in enumerate(list_block.bound_blocks):
                    # pass on the relevant sub-block raw value, should it be a ListBlock
                    try:
                        if isinstance(raw_value, list):
                            block_raw_value = raw_value[index]
                        elif isinstance(raw_value, dict):
                            block_raw_value = raw_value["value"][index]
                        else:
                            block_raw_value = None
                    except (IndexError, KeyError):
                        block_raw_value = None

                    segments.extend(
                        segment.wrap(block.id)
                        for segment in self.handle_block(
                            block.block, block.value, raw_value=block_raw_value
                        )
                    )
        return segments

    def handle_stream_block(self, stream_block):
        segments = []

        for index, block in enumerate(stream_block):
            raw_data = stream_block.raw_data[index]
            segments.extend(
                segment.wrap(block.id)
                for segment in self.handle_block(
                    block.block, block.value, raw_value=raw_data
                )
            )

        return segments


def extract_segments(instance):
    """
    Extracts segments from the given model instance.

    Args:
        instance (Model): The model instance to extract segments from.

    Returns:
        list[StringSegmentValue, TemplateSegmentValue, RelatedObjectSegmentValue, or OverridableSegmentValue]: The
            segment values that have been extracted.
    """
    segments = []

    for translatable_field in get_translatable_fields(instance.__class__):
        field = translatable_field.get_field(instance.__class__)
        is_translatable = translatable_field.is_translated(instance)
        is_synchronized = translatable_field.is_synchronized(instance)
        is_overridable = translatable_field.is_overridable(instance)
        extract_overridables = is_synchronized and is_overridable

        if hasattr(field, "get_translatable_segments"):
            if is_translatable:
                segments.extend(
                    segment.wrap(field.name)
                    for segment in field.get_translatable_segments(
                        field.value_from_object(instance)
                    )
                )

        elif isinstance(field, StreamField):
            if is_translatable:
                segments.extend(
                    segment.wrap(field.name)
                    for segment in StreamFieldSegmentExtractor(
                        field, include_overridables=extract_overridables
                    ).handle_stream_block(field.value_from_object(instance))
                )

        elif isinstance(field, RichTextField):
            value = field.value_from_object(instance)

            if is_translatable:
                template, strings = extract_strings(value)

                # Find all unique href values
                hrefs = set()
                for _string, attrs in strings:
                    for tag_attrs in attrs.values():
                        if "href" in tag_attrs:
                            hrefs.add(tag_attrs["href"])

                field_segments = (
                    [TemplateSegmentValue("", "html", template, len(strings))]
                    + [
                        StringSegmentValue("", string, attrs=attrs)
                        for string, attrs in strings
                    ]
                    + [
                        OverridableSegmentValue(quote_path_component(href), href)
                        for href in sorted(hrefs)
                    ]
                )

                segments.extend(segment.wrap(field.name) for segment in field_segments)

            if extract_overridables:
                pass  # TODO: Extract images and links

        elif isinstance(field, (models.TextField, models.CharField)):
            if not field.choices:
                value = field.value_from_object(instance)

                if value is None:
                    continue

                if is_translatable:
                    segments.append(StringSegmentValue(field.name, value))

                elif extract_overridables:
                    segments.append(OverridableSegmentValue(field.name, value))

        elif isinstance(field, models.ForeignKey):
            if is_translatable:
                if not issubclass(field.related_model, TranslatableMixin):
                    raise ImproperlyConfigured(
                        f"The foreign key `{field.model._meta.app_label}.{field.model.__name__}.{field.name}` was registered as a translatable "
                        f"field but the model it points to `{field.related_model._meta.app_label}.{field.related_model.__name__}` is not translatable"
                    )

                related_instance = getattr(instance, field.name)

                if related_instance:
                    segments.append(
                        RelatedObjectSegmentValue.from_instance(
                            field.name, related_instance
                        )
                    )

            elif extract_overridables:
                related_instance = getattr(instance, field.name)

                if related_instance:
                    segments.append(
                        OverridableSegmentValue(field.name, related_instance.pk)
                    )
        elif (
            isinstance(field, models.ManyToOneRel)
            and isinstance(field.remote_field, ParentalKey)
            and issubclass(field.related_model, TranslatableMixin)
        ):
            manager = getattr(instance, field.name)

            if is_translatable:
                for child_instance in manager.all():
                    segments.extend(
                        segment.wrap(str(child_instance.translation_key)).wrap(
                            field.name
                        )
                        for segment in extract_segments(child_instance)
                    )

            elif extract_overridables:
                pass  # TODO

    class Counter:
        def __init__(self):
            self.value = 0

        def next(self):
            self.value += 1
            return self.value

    counter = Counter()

    return [
        segment.with_order(counter.next())
        for segment in segments
        if not segment.is_empty()
    ]



from __future__ import annotations

from urllib.parse import urlencode

from django.contrib.admin.utils import quote
from django.contrib.auth.models import Permission
from django.db.models import Q
from django.shortcuts import redirect
from django.urls import include, path, reverse
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from django.views.i18n import JavaScriptCatalog
from wagtail import VERSION as WAGTAIL_VERSION
from wagtail import hooks
from wagtail.admin.action_menu import ActionMenuItem as PageActionMenuItem
from wagtail.admin.menu import MenuItem
from wagtail.admin.widgets import ListingButton
from wagtail.log_actions import LogFormatter
from wagtail.models import Locale, Page, TranslatableMixin
from wagtail.snippets.action_menu import ActionMenuItem as SnippetActionMenuItem
from wagtail.snippets.widgets import SnippetListingButton

# Import synctree so it can register its signal handler
from . import synctree  # noqa: F401
from .models import Translation, TranslationSource
from .views import (
    convert,
    edit_translation,
    report,
    snippets_api,
    submit_translations,
    update_translations,
)


@hooks.register("register_admin_urls")
def register_admin_urls():
    urls = [
        path(
            "jsi18n/",
            JavaScriptCatalog.as_view(packages=["wagtail_localize"]),
            name="javascript_catalog",
        ),
        path(
            "submit/page/<int:page_id>/",
            submit_translations.SubmitPageTranslationView.as_view(),
            name="submit_page_translation",
        ),
        path(
            "submit/snippet/<slug:app_label>/<slug:model_name>/<str:pk>/",
            submit_translations.SubmitSnippetTranslationView.as_view(),
            name="submit_snippet_translation",
        ),
        path(
            "update/<int:translation_source_id>/",
            update_translations.UpdateTranslationsView.as_view(),
            name="update_translations",
        ),
        path(
            "translate/<int:translation_id>/strings/<int:string_segment_id>/edit/",
            edit_translation.edit_string_translation,
            name="edit_string_translation",
        ),
        path(
            "translate/<int:translation_id>/overrides/<int:overridable_segment_id>/edit/",
            edit_translation.edit_override,
            name="edit_override",
        ),
        path(
            "translate/<int:translation_id>/pofile/download/",
            edit_translation.download_pofile,
            name="download_pofile",
        ),
        path(
            "translate/<int:translation_id>/pofile/upload/",
            edit_translation.upload_pofile,
            name="upload_pofile",
        ),
        path(
            "translate/<int:translation_id>/machine_translate/",
            edit_translation.machine_translate,
            name="machine_translate",
        ),
        path(
            "translate/<int:translation_id>/preview/",
            edit_translation.preview_translation,
            name="preview_translation",
        ),
        path(
            "translate/<int:translation_id>/preview/<str:mode>/",
            edit_translation.preview_translation,
            name="preview_translation",
        ),
        path(
            "translate/<int:translation_id>/disable/",
            edit_translation.stop_translation,
            name="stop_translation",
        ),
        path(
            "page/<int:page_id>/convert_to_alias/",
            convert.convert_to_alias,
            name="convert_to_alias",
        ),
        path(
            "reports/translations/",
            report.TranslationsReportView.as_view(),
            name="translations_report",
        ),
        path(
            "api/snippets/<slug:app_label>/<slug:model_name>/",
            snippets_api.SnippetViewSet.as_view({"get": "list"}),
            name="snippets_api_listing",
        ),
        path(
            "api/snippets/<slug:app_label>/<slug:model_name>/<str:pk>/",
            snippets_api.SnippetViewSet.as_view({"get": "retrieve"}),
            name="snippets_api_detail",
        ),
    ]

    if WAGTAIL_VERSION >= (6, 2):
        # Add a results-only view to add support for AJAX-based filtering
        urls.append(
            path(
                "reports/translations/results/",
                report.TranslationsReportView.as_view(results_only=True),
                name="translations_report_results",
            ),
        )

    return [
        path(
            "localize/",
            include(
                (urls, "wagtail_localize"),
                namespace="wagtail_localize",
            ),
        )
    ]


@hooks.register("register_permissions")
def register_submit_translation_permission():
    return Permission.objects.filter(
        content_type__app_label="wagtail_localize", codename="submit_translation"
    )


def page_listing_more_buttons(page: Page, user, view_name=None, next_url=None):
    if not page.is_root() and user.has_perm("wagtail_localize.submit_translation"):
        # If there's at least one locale that we haven't translated into yet, show "Translate this page" button
        has_locale_to_translate_to = Locale.objects.exclude(
            id__in=page.get_translations(inclusive=True)
            .exclude(alias_of__isnull=False)
            .values_list("locale_id", flat=True)
        ).exists()

        if has_locale_to_translate_to:
            # note: if this is an alias page, we want to use the parent page id
            url = reverse(
                "wagtail_localize:submit_page_translation",
                args=[page.alias_of_id or page.id],
            )

            yield ListingButton(
                _("Translate this page"),
                url,
                priority=60,
                icon_name="wagtail-localize-language",
            )

        # If the page is the source for translations, show "Sync translated pages" button
        source = TranslationSource.objects.get_for_instance_or_none(page)
        if source is not None and source.translations.filter(enabled=True).exists():
            url = reverse("wagtail_localize:update_translations", args=[source.id])
            if next_url is not None:
                url += "?" + urlencode({"next": next_url})

            yield ListingButton(
                _("Sync translated pages"), url, priority=60, icon_name="resubmit"
            )


hooks.register("register_page_header_buttons", page_listing_more_buttons)
hooks.register("register_page_listing_more_buttons", page_listing_more_buttons)


@hooks.register("register_snippet_listing_buttons")
def register_snippet_listing_buttons(snippet, user, next_url=None):
    model = type(snippet)

    if issubclass(model, TranslatableMixin) and user.has_perm(
        "wagtail_localize.submit_translation"
    ):
        # If there's at least one locale that we haven't translated into yet, show "Translate" button
        has_locale_to_translate_to = Locale.objects.exclude(
            id__in=snippet.get_translations(inclusive=True).values_list(
                "locale_id", flat=True
            )
        ).exists()

        if has_locale_to_translate_to:
            url = reverse(
                "wagtail_localize:submit_snippet_translation",
                args=[model._meta.app_label, model._meta.model_name, quote(snippet.pk)],
            )

            yield SnippetListingButton(
                _("Translate"),
                url,
                attrs={
                    "aria-label": _("Translate '%(title)s'") % {"title": str(snippet)}
                },
                icon_name="wagtail-localize-language",
                priority=100,
            )

        # If the snippet is the source for translations, show "Sync translated snippets" button
        source = TranslationSource.objects.get_for_instance_or_none(snippet)
        if source is not None and source.translations.filter(enabled=True).exists():
            url = reverse("wagtail_localize:update_translations", args=[source.id])
            if next_url is not None:
                url += "?" + urlencode({"next": next_url})

            yield SnippetListingButton(
                _("Sync translated snippets"),
                url,
                attrs={
                    "aria-label": _("Sync translations of '%(title)s'")
                    % {"title": str(snippet)}
                },
                icon_name="resubmit",
                priority=106,
            )


@hooks.register("before_edit_page")
def before_edit_page(request, page):
    # If the page is an alias of a page in another locale, override the edit page so that we can show a "Translate this page" option
    if page.alias_of and page.alias_of.locale_id != page.locale_id:
        return edit_translation.edit_translatable_alias_page(request, page)

    # Check if the user has clicked the "Start Synced translation" menu item
    if request.method == "POST":
        if "localize-restart-translation" in request.POST:
            try:
                translation = Translation.objects.get(
                    source__object_id=page.translation_key,
                    target_locale_id=page.locale_id,
                    enabled=False,
                )
            except Translation.DoesNotExist:
                pass
            else:
                return edit_translation.restart_translation(request, translation, page)
        elif "localize-convert-to-alias" in request.POST:
            return redirect(
                reverse("wagtail_localize:convert_to_alias", args=[page.id])
            )

    # Overrides the edit page view if the page is the target of a translation
    try:
        translation = Translation.objects.get(
            source__object_id=page.translation_key,
            target_locale_id=page.locale_id,
            enabled=True,
        )
        return edit_translation.edit_translation(request, translation, page)

    except Translation.DoesNotExist:
        pass


class RestartTranslationPageActionMenuItem(PageActionMenuItem):
    label = gettext_lazy("Start Synced translation")
    name = "localize-restart-translation"
    icon_name = "rotate"
    classname = "action-secondary"

    def is_shown(self, context):
        # Only show this menu item on the edit view where there was a previous translation record
        if context["view"] != "edit":
            return False

        return Translation.objects.filter(
            source__object_id=context["page"].translation_key,
            target_locale_id=context["page"].locale_id,
            enabled=False,
        ).exists()


@hooks.register("register_page_action_menu_item")
def register_restart_translation_page_action_menu_item():
    return RestartTranslationPageActionMenuItem(order=0)


class ConvertToAliasPageActionMenuItem(PageActionMenuItem):
    label = gettext_lazy("Convert to alias page")
    name = "localize-convert-to-alias"
    icon_name = "wagtail-localize-convert"
    classname = "action-secondary"

    def is_shown(self, context):
        # Only show this menu item on the edit view where there was a previous translation record
        if context["view"] != "edit":
            return False
        page = context["page"]
        try:
            return (
                page.alias_of_id is None
                and Page.objects.filter(
                    ~Q(pk=page.pk),
                    translation_key=page.translation_key,
                    locale_id=TranslationSource.objects.get(
                        object_id=page.translation_key,
                        specific_content_type=page.content_type_id,
                        translations__target_locale=page.locale_id,
                    ).locale_id,
                ).exists()
            )
        except TranslationSource.DoesNotExist:
            return False


@hooks.register("register_page_action_menu_item")
def register_convert_back_to_alias_page_action_menu_item():
    return ConvertToAliasPageActionMenuItem(order=0)


@hooks.register("before_edit_snippet")
def before_edit_snippet(request, instance):
    if isinstance(instance, TranslatableMixin):
        # Check if the user has clicked the "Start Synced translation" menu item
        if request.method == "POST" and "localize-restart-translation" in request.POST:
            try:
                translation = Translation.objects.get(
                    source__object_id=instance.translation_key,
                    target_locale_id=instance.locale_id,
                    enabled=False,
                )
            except Translation.DoesNotExist:
                pass
            else:
                return edit_translation.restart_translation(
                    request, translation, instance
                )

        # Overrides the edit snippet view if the snippet is translatable and the target of a translation
        try:
            translation = Translation.objects.get(
                source__object_id=instance.translation_key,
                target_locale_id=instance.locale_id,
                enabled=True,
            )
            return edit_translation.edit_translation(request, translation, instance)

        except Translation.DoesNotExist:
            pass


class RestartTranslationSnippetActionMenuItem(SnippetActionMenuItem):
    label = gettext_lazy("Start Synced translation")
    name = "localize-restart-translation"
    icon_name = "rotate"
    classname = "action-secondary"

    def is_shown(self, context):
        # Only show this menu item on the edit view where there was a previous translation record
        if context["view"] != "edit":
            return False

        if not issubclass(context["model"], TranslatableMixin):
            return False

        return Translation.objects.filter(
            source__object_id=context["instance"].translation_key,
            target_locale_id=context["instance"].locale_id,
            enabled=False,
        ).exists()


@hooks.register("register_snippet_action_menu_item")
def register_restart_translation_snippet_action_menu_item(model):
    return RestartTranslationSnippetActionMenuItem(order=0)


class TranslationsReportMenuItem(MenuItem):
    def is_shown(self, request):
        return True


@hooks.register("register_reports_menu_item")
def register_wagtail_localize2_report_menu_item():
    return TranslationsReportMenuItem(
        _("Translations"),
        reverse("wagtail_localize:translations_report"),
        icon_name="site",
        order=9000,
    )


@hooks.register("register_log_actions")
def wagtail_localize_log_actions(actions):
    @actions.register_action("wagtail_localize.convert_to_alias")
    class ConvertToAliasActionFormatter(LogFormatter):
        label = gettext_lazy("Convert page to alias")

        def format_message(self, log_entry):
            try:
                return _(
                    "Converted page '%(title)s' to an alias of the translation source page '%(source_title)s'"
                ) % {
                    "title": log_entry.data["page"]["title"],
                    "source_title": log_entry.data["source"]["title"],
                }
            except KeyError:
                return _("Converted page to an alias of the translation source page")


@hooks.register("register_icons")
def register_icons(icons):
    return icons + [
        # icon id "wagtail-localize-convert" (which translates to `.icon-wagtail-localize-convert`)
        "wagtail_localize/icons/wagtail-localize-convert.svg",
        # icon id "wagtail-localize-language" (which translates to `.icon-wagtail-localize-language`)
        "wagtail_localize/icons/wagtail-localize-language.svg",
    ]



from django.contrib.contenttypes.models import ContentType

from wagtail_localize.strings import StringValue


class BaseValue:
    def __init__(self, path, order=0):
        self.path = path
        self.order = order

    def clone(self):
        """
        Clones this segment. Must be overridden in subclass.
        """
        raise NotImplementedError

    def with_order(self, order):
        """
        Sets the order of this segment.
        """
        clone = self.clone()
        clone.order = order
        return clone

    def wrap(self, base_path):
        """
        Appends a component to the beginning of the path.

        For example:

        >>> s = StringSegmentValue("field", "foo")
        >>> s.wrap("wrapped")
        StringSegmentValue('wrapped.field', 'foo')
        """
        new_path = base_path

        if self.path:
            new_path += "." + self.path

        clone = self.clone()
        clone.path = new_path
        return clone

    def unwrap(self):
        """
        Pops a component from the beginning of the path. Reversing .wrap().

        For example:

        >>> s = StringSegmentValue("wrapped.field", "foo")
        >>> s.unwrap()
        'wrapped', StringSegmentValue('field', 'foo')
        """
        first_component, *remaining_components = self.path.split(".")
        new_path = ".".join(remaining_components)

        clone = self.clone()
        clone.path = new_path
        return first_component, clone


class StringSegmentValue(BaseValue):
    """
    Represents a translatable string segment.

    Attributes:
        path (str): The content path of the segment.
        string (StringValue): the value of the segment.
        attrs (dict): A dict of HTML attributes that were stripped out of the string.
        order (int): The index that this segment appears on a page.
    """

    def __init__(self, path, string, attrs=None, **kwargs):
        """
        Initialises a new StringSegmentValue.

        Args:
            path (str): The content path of the segment.
            string (StringValue): the value of the segment.
            attrs (dict, optional): A dict of HTML attributes that were stripped out of the string.
            order (int, optional): The index that this segment appears on a page.
        """
        if isinstance(string, str):
            string = StringValue.from_plaintext(string)

        elif isinstance(string, StringValue):
            pass

        else:
            raise TypeError(
                f"`string` must be either a `StringValue` or a `str`. Got `{type(string).__name__}`"
            )

        self.string = string
        self.attrs = attrs or None

        super().__init__(path, **kwargs)

    def clone(self):
        """
        Makes an exact copy of this StringSegmentValue.

        Instead of mutating SegmentValue classes in place, it's recommended to clone them first.

        Returns:
            StringSegmentValue: The new segment value that's a copy of this one.
        """
        return StringSegmentValue(
            self.path, self.string, attrs=self.attrs, order=self.order
        )

    @classmethod
    def from_source_html(cls, path, html, **kwargs):
        """
        Initialises a StringSegmentValue from a HTML string.

        Args:
            path (str): The content path of the segment.
            html (str): The HTML value of the segment.
            order (int, optional): The index that this segment appears on a page.
        """
        string, attrs = StringValue.from_source_html(html)
        return cls(path, string, attrs=attrs, **kwargs)

    def render_text(self):
        """
        Returns a plain text representation of the segment value.

        Note: If the segment value was created from HTML, this strips out HTML tags.

        Returns:
            str: The text representation of the segment value.
        """
        return self.string.render_text()

    def render_html(self):
        """
        Returns a HTML representation of the segment value.

        Note: If the segment value was created from plain text, this escapes special characters.

        Returns:
            str: The HTML representation of the segment value.
        """
        return self.string.render_html(self.attrs)

    def is_empty(self):
        """
        Returns True if the StringValue is blank.

        Returns:
            boolean: True if the StringValue is blank.
        """
        return self.string.data in ["", None]

    def __eq__(self, other):
        return (
            isinstance(other, StringSegmentValue)
            and self.path == other.path
            and self.string == other.string
            and self.attrs == other.attrs
        )

    def __repr__(self):
        return f"<StringSegmentValue {self.path} '{self.render_html()}'>"


class TemplateSegmentValue(BaseValue):
    """
    Represents an HTML template used to recombine the values of RichTextField/Blocks.

    Attributes:
        path (str): The content path of the segment.
        format (str): The format of the template (eg, 'html').
        template (str): The template.
        string_count (int): The number of translatablle string segments that were extracted from the template.
        order (int): The index that this segment appears on a page.
    """

    def __init__(self, path, format, template, string_count, **kwargs):
        """
        Initialises a new TemplateSegmentValue.

        Args:
            path (str): The content path of the segment.
            format (str): The format of the template (eg, 'html').
            template (str): The template.
            string_count (int): The number of translatablle string segments that were extracted from the template.
            order (int, optional): The index that this segment appears on a page.
        """
        self.format = format
        self.template = template
        self.string_count = string_count

        super().__init__(path, **kwargs)

    def clone(self):
        """
        Makes an exact copy of this TemplateSegmentValue.

        Instead of mutating SegmentValue classes in place, it's recommended to clone them first.

        Returns:
            TemplateSegmentValue: The new segment value that's a copy of this one.
        """
        return TemplateSegmentValue(
            self.path, self.format, self.template, self.string_count, order=self.order
        )

    def is_empty(self):
        """
        Returns True if the template is blank.

        Returns:
            boolean: True if the template is blank.
        """
        return self.template in ["", None]

    def __eq__(self, other):
        return (
            isinstance(other, TemplateSegmentValue)
            and self.path == other.path
            and self.format == other.format
            and self.template == other.template
            and self.string_count == other.string_count
        )

    def __repr__(self):
        return f"<TemplateSegmentValue {self.path} format:{self.format} {self.string_count} segments>"


class RelatedObjectSegmentValue(BaseValue):
    """
    Represents a reference to a foreign translatable object.

    Attributes:
        path (str): The content path of the segment.
        content_type (ContentType): The content type of the base model of the foreign object.
        translation_key (UUID): The value of the foreign object's `translation_key` field.
        order (int): The index that this segment appears on a page.
    """

    def __init__(self, path, content_type, translation_key, **kwargs):
        """
        Initialises a new RelatedObjectSegmentValue.

        Args:
            path (str): The content path of the segment.
            content_type (ContentType): The content type of the base model of the foreign object.
            translation_key (UUID): The value of the foreign object's `translation_key` field.
            order (int, optional): The index that this segment appears on a page.
        """
        self.content_type = content_type
        self.translation_key = translation_key

        super().__init__(path, **kwargs)

    @classmethod
    def from_instance(cls, path, instance):
        """
        Initialises a new RelatedObjectSegmentValue from a model instance.

        Note: This class only records the type and translation key of the instance. So you will need to save the
        locale separately if you need to get this same instance back later.

        Args:
            path (str): The content path of the segment.
            instance (Model): An instance of the translatable object that needs to be referenced.

        Raises:
            AttributeError: If the instance is not translatable.

        Returns:
            RelatedObjectSegmentValue: The new segment value.
        """
        model = instance.get_translation_model()
        return cls(
            path, ContentType.objects.get_for_model(model), instance.translation_key
        )

    def get_instance(self, locale):
        """
        Gets an instance of the referenced translatable object for the given locale.

        Raises:
            Model.DoesNotExist: If there isn't an instance of the translatable object in the given locale.

        Returns:
            Model: The instance.
        """
        from ..models import pk

        return self.content_type.get_object_for_this_type(
            translation_key=self.translation_key, locale_id=pk(locale)
        )

    def clone(self):
        """
        Makes an exact copy of this RelatedObjectSegmentValue.

        Instead of mutating SegmentValue classes in place, it's recommended to clone them first.

        Returns:
            RelatedObjectSegmentValue: The new segment value that's a copy of this one.
        """
        return RelatedObjectSegmentValue(
            self.path, self.content_type, self.translation_key, order=self.order
        )

    def is_empty(self):
        """
        Returns True if the related object is null.

        Returns:
            boolean: True if the related object is null.
        """
        return self.content_type is None and self.translation_key is None

    def __eq__(self, other):
        return (
            isinstance(other, RelatedObjectSegmentValue)
            and self.path == other.path
            and self.content_type == other.content_type
            and self.translation_key == other.translation_key
        )

    def __repr__(self):
        return f"<RelatedObjectSegmentValue {self.path} {self.content_type} {self.translation_key}>"


class OverridableSegmentValue(BaseValue):
    """
    Represents a field that can be overridden.

    Attributes:
        path (str): The content path of the segment.
        data (any): The value of the field in the source. Must be JSON-serializable.
        order (int): The index that this segment appears on a page.
    """

    def __init__(self, path, data, **kwargs):
        """
        Initialises a new RelatedObjectSegmentValue.

        Args:
            path (str): The content path of the segment.
            data (any): The value of the field in the source. Must be JSON-serializable.
            order (int, optional): The index that this segment appears on a page.
        """
        self.data = data

        super().__init__(path, **kwargs)

    def clone(self):
        """
        Makes an exact copy of this OverridableSegmentValue.

        Instead of mutating SegmentValue classes in place, it's recommended to clone them first.

        Returns:
            OverridableSegmentValue: The new segment value that's a copy of this one.
        """
        return OverridableSegmentValue(self.path, self.data, order=self.order)

    def is_empty(self):
        """
        Returns True if the data is empty.

        Returns:
            boolean: True if the data is empty.
        """
        return self.data in ["", None]

    def __eq__(self, other):
        return (
            isinstance(other, OverridableSegmentValue)
            and self.path == other.path
            and self.data == other.data
        )

    def __repr__(self):
        return f"<OverridableSegmentValue {self.path} '{self.data}'>"



# This file is heavily inspired by django.utils.version


def get_version(version):
    """Return a PEP 440-compliant version number from VERSION."""
    version = get_complete_version(version)

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|rc}N - for alpha, beta, and rc releases

    main = get_main_version(version)

    sub = ""
    if version[3] != "final":
        mapping = {"alpha": "a", "beta": "b", "rc": "rc", "dev": ".dev"}
        sub = mapping[version[3]] + str(version[4])

    return main + sub


def get_main_version(version=None):
    """Return main version (X.Y[.Z]) from VERSION."""
    version = get_complete_version(version)
    parts = 2 if version[2] == 0 else 3
    return ".".join(str(x) for x in version[:parts])


def get_complete_version(version=None):
    """
    Return a tuple of the Wagtail version. If version argument is non-empty,
    check for correctness of the tuple provided.
    """
    if version is None:
        from wagtail_localize.version import VERSION as version
    else:
        if len(version) != 5:
            raise ValueError("wagtail_localize version number must be a 5-tuple")
        if version[3] not in ("dev", "alpha", "beta", "rc", "final"):
            raise ValueError(
                "wagtail_localize version status must be one of: dev, alpha, beta, rc, or final"
            )

    return version



# Generated by Django 3.2.4 on 2021-07-20 14:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wagtail_localize", "0014_remove_translation_source_last_updated_at"),
    ]

    operations = [
        migrations.AddField(
            model_name="translationcontext",
            name="field_path",
            field=models.TextField(default=""),
            preserve_default=False,
        ),
    ]



from urllib.parse import urlencode

from django.contrib.admin.utils import quote
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from wagtail.models import Locale, Page, TranslatableMixin

from wagtail_localize.models import TranslationSource


try:
    from wagtail_modeladmin.helpers import ButtonHelper, PageButtonHelper
    from wagtail_modeladmin.views import InspectView
except ImportError:
    from wagtail.contrib.modeladmin.helpers import ButtonHelper, PageButtonHelper
    from wagtail.contrib.modeladmin.views import InspectView


class TranslatableButtonHelper(ButtonHelper):
    def get_buttons_for_obj(self, obj, **kwargs):
        btns = super().get_buttons_for_obj(obj, **kwargs)
        user = self.request.user
        next_url = self.request.get_full_path()
        if isinstance(self.view, InspectView):
            classname = "button button-secondary"
        else:
            classname = "button button-secondary button-small"
        btns += list(get_translation_buttons(obj, user, next_url, classname))
        return btns


class TranslatablePageButtonHelper(TranslatableButtonHelper, PageButtonHelper):
    pass


def get_translation_buttons(obj, user, next_url=None, classname=""):
    """
    Generate listing buttons for translating/syncing objects in modeladmin.
    """
    model = type(obj)

    if issubclass(model, TranslatableMixin) and user.has_perm(
        "wagtail_localize.submit_translation"
    ):
        # If there's at least one locale that we haven't translated into yet, show "Translate" button
        if isinstance(obj, Page):
            has_locale_to_translate_to = Locale.objects.exclude(
                id__in=obj.get_translations(inclusive=True)
                .exclude(alias_of__isnull=False)
                .values_list("locale_id", flat=True)
            ).exists()
        else:
            has_locale_to_translate_to = Locale.objects.exclude(
                id__in=obj.get_translations(inclusive=True).values_list(
                    "locale_id", flat=True
                )
            ).exists()

        if has_locale_to_translate_to:
            url = reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=[model._meta.app_label, model._meta.model_name, quote(obj.pk)],
            )
            yield {
                "url": url,
                "label": _("Translate"),
                "classname": classname,
                "title": _("Translate"),
            }

        # If the object is the source for translations, show "Sync translated" button
        source = TranslationSource.objects.get_for_instance_or_none(obj)
        if source is not None and source.translations.filter(enabled=True).exists():
            url = reverse("wagtail_localize:update_translations", args=[source.id])
            if next_url is not None:
                url += "?" + urlencode({"next": next_url})

            yield {
                "url": url,
                "label": _("Sync translated %(model_name)s")
                % {"model_name": obj._meta.verbose_name_plural},
                "classname": classname,
                "title": _("Sync translated %(model_name)s")
                % {"model_name": obj._meta.verbose_name_plural},
            }



class BaseMachineTranslator:
    display_name = "Unknown"

    def __init__(self, options):
        self.options = options

    def translate(self, source_locale, target_locale, strings):
        raise NotImplementedError

    def can_translate(self, source_locale, target_locale):
        return False



# Generated by Django 4.2.9 on 2024-01-06 13:20

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wagtail_localize", "0015_translationcontext_field_path"),
    ]

    operations = [
        migrations.RenameField(
            model_name="translationlog",
            old_name="page_revision",
            new_name="revision",
        ),
    ]



# Generated by Django 3.2.6 on 2021-09-14 08:53

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wagtail_localize", "0013_translationsource_schema_version"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="translation",
            name="source_last_updated_at",
        ),
    ]



# Generated by Django 3.0.8 on 2020-08-12 18:34

import django.db.models.deletion

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("wagtail_localize", "0007_stringtranslation_type_and_tool_name"),
    ]

    operations = [
        migrations.AddField(
            model_name="stringtranslation",
            name="last_translated_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]
